/*
  mkhomedir.c
  Matlib 2010-01-19
*/

#include "defs.h"
#include "strings.c"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>
#include <dirent.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <pwd.h>
#include <grp.h>

#include <arguments.h>



#define CHECK_ERROR(__COND__, ...) \
  if (__COND__) \
  { \
    fprintf (stderr, __VA_ARGS__); \
    exit (EXIT_FAILURE); \
  };
  
int be_quiet (const char *format, ...)
{
  return 0;
}
int (*verbose) (const char *, ...) = be_quiet;




ARGUMENT_CALLBACK (argument_uid);
ARGUMENT_CALLBACK (argument_gid);
ARGUMENT_CALLBACK (argument_overwr);
ARGUMENT_CALLBACK (argument_verbose);

struct Argument arguments [] =
{
  {'d', "homedir",   argument_string,  ARGUMENT_F_ONCE | ARGUMENT_F_PAR, DEFAULT_HOME},
  {'H', "forcehome", argument_bool,    ARGUMENT_F_ONCE,                  NULL},
  {'s', "skel",      argument_string,  ARGUMENT_F_ONCE | ARGUMENT_F_PAR, DEFAULT_SKEL},
  {'u', "uid",       argument_uid,     ARGUMENT_F_ONCE | ARGUMENT_F_PAR, (void *) -1},
  {'g', "gid",       argument_gid,     ARGUMENT_F_ONCE | ARGUMENT_F_PAR, (void *) -1},
  {'f', "overwrite", argument_overwr,  ARGUMENT_F_ONCE,                  NULL},
  {'v', "verbose",   argument_verbose, ARGUMENT_F_ONCE,                  NULL},
  { }
};
#define ARG_HOME    0
#define ARG_FHOME   1
#define ARG_SKEL    2
#define ARG_UID     3
#define ARG_GID     4
#define ARG_OVERWR  5
#define ARG_VERBOSE 6

#define HOME  ((char *) (arguments [ARG_HOME].value))
#define FHOME (arguments [ARG_FHOME].value != NULL)
#define SKEL  ((char *) (arguments [ARG_SKEL].value))
#define UID   ((long int) (arguments [ARG_UID].value));
#define GID   ((long int) (arguments [ARG_GID].value));
#define OVERWRITE (arguments [ARG_OVERWR].value != NULL)



ARGUMENT_CALLBACK (argument_uid)
{
  struct passwd *p;
  long int v;
  char *tail;
  
  p = getpwnam (parameter);
  if (p == NULL)
  {
    v = strtol (parameter, &tail, 0);
    if (*tail == 0)
    {
      arg -> value = (void *) v;
      return;
    }
    arguments_error_freeform ("No such user \"%s\"", parameter);
  }
  arg -> value = (void *) ((long int) (p -> pw_uid));
}

ARGUMENT_CALLBACK (argument_gid)
{
  struct group *g;
  long int v;
  char *tail;
  
  g = getgrnam (parameter);
  if (g == NULL)
  {
    v = strtol (parameter, &tail, 0);
    if (*tail == 0)
    {
      arg -> value = (void *) v;
      return;
    }
    arguments_error_freeform ("No such group \"%s\"", parameter);
  }
  arg -> value = (void *) ((long int) (g -> gr_gid));
}

ARGUMENT_CALLBACK (argument_overwr)
{
#if ALLOW_F == 0
  if (getuid () == geteuid ())
    arg -> value = (void *) 1;
  else
    arguments_error (arg, "may only be used when real and effective UIDs match");
#elif ALLOW_F == 1
  if (getgid () == getegid ())
    arg -> value = (void *) 1;
  else
    arguments_error (arg, "may only be used when real and effective GIDs match");
#else
  arg -> value = (void *) 1;
#endif
}

ARGUMENT_CALLBACK (argument_verbose)
{
  verbose = printf;
}

void clean_arguments ()
{
  int i;
  for (i = 0; arguments [i].shrt != 0; i ++)
    if ((i != ARG_FHOME) && (i != ARG_VERBOSE))
      arguments [i].flags &= ~ARGUMENT_F_OCC;
}





/* Get UID and GID values and user's home directory from password 
   database */

int user_uid, user_gid;

void set_ids (char *user, String *home)
{
  struct passwd *p;
  user_uid = UID;
  user_gid = GID;
  if ((user_uid == -1) || (user_gid == -1) || (!FHOME))
  {
    p = getpwnam (user);
    CHECK_ERROR (p == NULL, "getpwnam: %s: user not found in password database\n", user);
    if (user_uid == -1)
      user_uid = p -> pw_uid;
    if (user_gid == -1)
      user_gid = p -> pw_gid;
  }
  if (FHOME || (p -> pw_dir == NULL))
    string_fmt (home, "%s/%s", HOME, user);
  else
    string_fmt (home, "%s", p -> pw_dir);
  verbose ("User \"%s\" (uid=%d, gid=%d) with home \"%s\"\n", user,
     user_uid, user_gid, home -> data);
}






/* Utility functions.
   check_directory: check if skel is a directory (or a symbolic 
                    link to directory) 
   replace_dir:     when overwriting -- reuse already existing
                    directory or unlink the item otherwise */

void check_directory (char *name, struct stat *s)
{
  CHECK_ERROR (stat (name, s) != 0, "%s: stat: %s\n", name, strerror (errno));
  CHECK_ERROR (!S_ISDIR (s -> st_mode), "%s is not a directory\n", name);
}

void copy_file (char *from, char *to)
{
  char buffer [1024];
  int i, o;
  ssize_t r, w, p;
  i = open (from, O_RDONLY | O_NOCTTY);
  CHECK_ERROR (i == -1, "%s: open (r): %s\n", from, strerror (errno));
  o = open (to, O_WRONLY | O_CREAT | O_EXCL, 0700);
  if ((o == -1) && (errno == EEXIST) && OVERWRITE)
  {
    verbose ("    overwriting\n");
    CHECK_ERROR (unlink (to) != 0, "%s: unlink: %s", to, strerror (errno));
    o = open (to, O_WRONLY | O_CREAT | O_EXCL, 0700);
  }
  CHECK_ERROR (o == -1, "%s: open (w): %s\n", to, strerror (errno));
loop:
  r = read (i, buffer, 1024);
  if (r == 0)
  {
    close (i);
    close (o);
    return;
  }
  CHECK_ERROR (r == -1, "%s: read: %s\n", from, strerror (errno));
  p = 0;
  do
  {
    w = write (o, buffer + p, r - p);
    CHECK_ERROR (w == -1, "%s: write: %s\n", from, strerror (errno));
    p += w;
  }
  while (p < r);
  goto loop;
}

int replace_dir (char *name, mode_t mode)
{
  struct stat s;
  verbose ("    overwriting\n");
  CHECK_ERROR (lstat (name, &s) != 0, "%s: lstat: %s\n", name, strerror (errno));
  if (S_ISDIR (s.st_mode))
  {
    CHECK_ERROR (chmod (name, mode) != 0, "%s: chmod (0%o): %s\n", name, mode, strerror (errno));
    return 0;
  }
  CHECK_ERROR (unlink (name), "%s: unlink: %s\n", name, strerror (errno));
  return mkdir (name, mode);
}





/* Main procedures here. process_directory recursively calls cp for each
   entry */

void cp (String *from, String *to, struct stat *s);

void process_directory (String *from, String *to)
{
  DIR *d;
  struct dirent *de;
  size_t fl, tl;
  struct stat s;
  
  fl = from -> l;
  tl = to -> l;
  d = opendir (from -> data);
  CHECK_ERROR (d == NULL, "%s: opendir: %s\n", from -> data, strerror (errno));
loop:
  de = readdir (d);
  if (de == NULL)
  {
    CHECK_ERROR (errno == EBADF, "%s: readdir: %s\n", from -> data, strerror (errno));
    closedir (d);
    from -> data [from -> l] = 0;
    to -> data [to -> l] = 0;
    return;
  }
  if ((strcmp (de -> d_name, ".") != 0) && (strcmp (de -> d_name, "..") != 0))
  {
    string_afmt (from, "/%s", de -> d_name);
    string_afmt (to, "/%s", de -> d_name);
    CHECK_ERROR (lstat (from -> data, &s) != 0, "%s: lstat: %s\n", from -> data, strerror (errno));
    cp (from, to, &s);
    from -> l = fl;
    to -> l = tl;
  }
  goto loop;
}


void cp (String *from, String *to, struct stat *s)
{
  int uid, gid, mode, e;
  
  if (s -> st_uid == 0)
    uid = user_uid;
  else
    uid = s -> st_uid;
  if (s -> st_gid == 0)
    gid = user_gid;
  else
    gid = s -> st_gid;
  if (S_ISDIR (s -> st_mode))
    mode = (s -> st_mode) & (~S_ISUID);
  else
    mode = (s -> st_mode) & (~(S_ISUID | S_ISGID | S_ISVTX));
  verbose ("  \"%s\" (%d.%d 0%o)\n", to -> data,
     uid, gid, mode & 07777);
  switch (mode & S_IFMT)
  {
    case S_IFDIR:
      e = mkdir (to -> data, 0700);
      if ((e != 0) && (errno == EEXIST) && OVERWRITE)
        e = replace_dir (to -> data, 0700);;
      CHECK_ERROR (e != 0, "%s: mkdir: %s\n", to -> data, strerror (errno));
      process_directory (from, to);
      break;
    case S_IFCHR:
    case S_IFBLK:
    case S_IFSOCK:
    case S_IFIFO:
      CHECK_ERROR (mknod (to -> data, mode, s -> st_rdev) != 0, "%s: mknod: %s\n", to -> data, strerror (errno));
      break;
    case S_IFREG:
      copy_file (from -> data, to -> data);
      break;
    case S_IFLNK:
      new_string (link_data);
      CHECK_ERROR ((e = string_readlink (&link_data, from -> data)) != 0, "%s: readlink: %s\n", from -> data, strerror (e));
      CHECK_ERROR (symlink (to -> data, from -> data) != 0, "%s: symlink: %s\n", to -> data, strerror (errno));
      return;
    default:
      fprintf (stderr, "%s: Unknown node type: 0x%X\n", 
         from -> data, (mode) & S_IFMT);
      exit (EXIT_FAILURE);
  }
  
  CHECK_ERROR (chmod (to -> data, mode) != 0, "%s: chmod (0%o): %s\n", to -> data, mode, strerror (errno));
  CHECK_ERROR (chown (to -> data, uid, gid) != 0, "%s: chown (%d.%d): %s\n", to -> data, uid, gid, strerror (errno));
}


int main (int argc, char **argv)
{
  int non_option;
  struct stat s;
  non_option = 0;
loop:
  non_option += arguments_parse ("mkhomedir", "1", 
      "user user ... [OPTION]... user ...",
      "Creates home directory(s) for specified user(s)", 
      argc - non_option, argv + non_option, arguments);
  if (non_option == argc)
    return EXIT_SUCCESS;
  new_string (src);
  new_string (dst);
  set_ids (argv [non_option], &dst);
  check_directory (SKEL, &s);
  string_str (&src, SKEL);
  cp (&src, &dst, &s);
  clean_arguments ();
  goto loop;
}

